import { copy, flatMerge, isValidObject, StateObserver, equal, isFunction, notEqual, generateId, removeProperties, LogCodeManager, defineConfig, shared, } from '../internal';
export class State {
    constructor(agileInstance, initialValue, config = {}) {
        this.isSet = false;
        this.isPlaceholder = false;
        this.observers = {};
        this.sideEffects = {};
        this.isPersisted = false;
        this.watchers = {};
        config = defineConfig(config, {
            dependents: [],
            isPlaceholder: false,
        });
        this.agileInstance = () => agileInstance;
        this._key = config.key;
        this.observers['value'] = new StateObserver(this, {
            key: config.key,
            dependents: config.dependents,
        });
        this.initialStateValue = copy(initialValue);
        this._value = copy(initialValue);
        this.previousStateValue = copy(initialValue);
        this.nextStateValue = copy(initialValue);
        this.isPlaceholder = true;
        this.computeExistsMethod = (v) => {
            return v != null;
        };
        if (!config.isPlaceholder)
            this.set(initialValue, { overwrite: true });
    }
    set value(value) {
        this.set(value);
    }
    get value() {
        return copy(this._value);
    }
    set key(value) {
        this.setKey(value);
    }
    get key() {
        return this._key;
    }
    setKey(value) {
        const oldKey = this._key;
        this._key = value;
        for (const observerKey in this.observers)
            this.observers[observerKey]._key = value;
        return this;
    }
    set(value, config = {}) {
        config = defineConfig(config, {
            force: false,
        });
        const _value = isFunction(value)
            ? value(copy(this._value))
            : value;
        if (!this.hasCorrectType(_value)) {
            LogCodeManager.log(config.force ? '14:02:00' : '14:03:00', [
                typeof _value,
                this.valueType,
            ]);
            if (!config.force)
                return this;
        }
        this.observers['value'].ingestValue(_value, config);
        return this;
    }
    ingest(config = {}) {
        this.observers['value'].ingest(config);
        return this;
    }
    type(type) {
        const supportedTypes = ['String', 'Boolean', 'Array', 'Object', 'Number'];
        if (!supportedTypes.includes(type.name)) {
            LogCodeManager.log('14:03:01', [type]);
            return this;
        }
        this.valueType = type.name.toLowerCase();
        return this;
    }
    undo(config = {}) {
        this.set(this.previousStateValue, config);
        return this;
    }
    reset(config = {}) {
        this.set(this.initialStateValue, config);
        return this;
    }
    patch(targetWithChanges, config = {}) {
        config = defineConfig(config, {
            addNewProperties: true,
        });
        if (!isValidObject(this.nextStateValue, true)) {
            LogCodeManager.log('14:03:02');
            return this;
        }
        if (!isValidObject(targetWithChanges, true)) {
            LogCodeManager.log('00:03:01', ['TargetWithChanges', 'object']);
            return this;
        }
        if (Array.isArray(targetWithChanges) &&
            Array.isArray(this.nextStateValue)) {
            this.nextStateValue = [
                ...this.nextStateValue,
                ...targetWithChanges,
            ];
        }
        else {
            this.nextStateValue = flatMerge(this.nextStateValue, targetWithChanges, { addNewProperties: config.addNewProperties });
        }
        this.ingest(removeProperties(config, ['addNewProperties']));
        return this;
    }
    watch(keyOrCallback, callback) {
        const generateKey = isFunction(keyOrCallback);
        let _callback;
        let key;
        if (generateKey) {
            key = generateId();
            _callback = keyOrCallback;
        }
        else {
            key = keyOrCallback;
            _callback = callback;
        }
        if (!isFunction(_callback)) {
            LogCodeManager.log('00:03:01', ['Watcher Callback', 'function']);
            return this;
        }
        this.watchers[key] = _callback;
        return generateKey ? key : this;
    }
    removeWatcher(key) {
        delete this.watchers[key];
        return this;
    }
    hasWatcher(key) {
        return !!this.watchers[key];
    }
    onInaugurated(callback) {
        const watcherKey = 'InauguratedWatcherKey';
        this.watch(watcherKey, (value, key) => {
            callback(value, key);
            this.removeWatcher(watcherKey);
        });
        return this;
    }
    get exists() {
        return !this.isPlaceholder && this.computeExistsMethod(this.value);
    }
    computeExists(method) {
        if (!isFunction(method)) {
            LogCodeManager.log('00:03:01', ['Compute Exists Method', 'function']);
            return this;
        }
        this.computeExistsMethod = method;
        return this;
    }
    computeValue(method) {
        if (!isFunction(method)) {
            LogCodeManager.log('00:03:01', ['Compute Value Method', 'function']);
            return this;
        }
        this.computeValueMethod = method;
        this.set(this.nextStateValue);
        return this;
    }
    is(value) {
        return equal(value, this.value);
    }
    isNot(value) {
        return notEqual(value, this.value);
    }
    invert() {
        switch (typeof this.nextStateValue) {
            case 'boolean':
                this.set(!this.nextStateValue);
                break;
            case 'object':
                if (Array.isArray(this.nextStateValue))
                    this.set(this.nextStateValue.reverse());
                break;
            case 'string':
                this.set(this.nextStateValue.split('').reverse().join(''));
                break;
            case 'number':
                this.set((this.nextStateValue * -1));
                break;
            default:
                LogCodeManager.log('14:03:04', [typeof this.nextStateValue]);
        }
        return this;
    }
    addSideEffect(key, callback, config = {}) {
        config = defineConfig(config, {
            weight: 10,
        });
        if (!isFunction(callback)) {
            LogCodeManager.log('00:03:01', ['Side Effect Callback', 'function']);
            return this;
        }
        this.sideEffects[key] = {
            callback: callback,
            weight: config.weight,
        };
        return this;
    }
    removeSideEffect(key) {
        delete this.sideEffects[key];
        return this;
    }
    hasSideEffect(key) {
        return !!this.sideEffects[key];
    }
    hasCorrectType(value) {
        if (!this.valueType)
            return true;
        const type = typeof value;
        return type === this.valueType;
    }
    getPersistableValue() {
        return this._value;
    }
}
export function createState(initialValue, config = {}) {
    config = defineConfig(config, {
        agileInstance: shared,
    });
    return new State(config.agileInstance, initialValue, removeProperties(config, ['agileInstance']));
}
